import argparse, sys, os
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk
import bz2, re
import nltk

#<doc id="12908" url="https://en.wikipedia.org/wiki?curid=12908" title="Global warming potential">
ID_REGEX = re.compile(' id="([^"]*)"')
URL_REGEX = re.compile(' url="([^"]*)"')
TITLE_REGEX = re.compile(' title="([^"]*)"')

class SentenceIteratorWrapper:
    def __init__(self, filename, index_name):
        self.filename = filename
        self.index_name = index_name
        self.num_sentences_read = 0

    def iterator(self, start_id):
        if self.filename.endswith(".bz2"):
            f = bz2.open(self.filename, "rt", encoding='utf-8')
        else:
            f = open(self.filename, encoding='utf-8')
        text_buffer = []
        title = None
        for line in f:
            if line.startswith('<doc id="'):
                title_match = TITLE_REGEX.search(line)
                title = title_match.group(1)
                #id_match = ID_REGEX.search(line)
                #url_match = URL_REGEX.search(line)
                #if not (id_match and url_match and title_match):
                #    raise ValueError("Unexpected article header: " + line)
                #docid = id_match.group(1)
                #url = url_match.group(1)
            elif line.startswith("</doc>"):
                text = "".join(text_buffer)
                sentences = nltk.sent_tokenize(text)
                for sentence in sentences:
                    whitespaces = sentence.count(" ")
                    if whitespaces > 4 and whitespaces < 40: # Only "standard-length" sentences.
                        self.num_sentences_read += 1
                        yield {
                            "_index": self.index_name,
                            "_type": "sentence",
                            "_source": {
                                "title": title,
                                "text": sentence,
                                "id": self.num_sentences_read + start_id - 1
                            }
                        }
                text_buffer = []
            else:
                text_buffer.append(line)

def files_in_dir(base_directory):
    filelist = []
    for path,dirs,files in os.walk(base_directory):
        for filename in files:
            filelist.append(os.path.join(path,filename))
    return filelist

def main(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--directory', required = True)
    opts = parser.parse_args()

    es = Elasticsearch()

    # Delete index if it already exists
    if es.indices.exists(index="wiki-sentences"):
        es.indices.delete(index="wiki-sentences")

    files = files_in_dir(opts.directory)
    sentence_id = 0
    for i in range(len(files)):
        print("\rFile %d out of %d" % (i+1,len(files)), end = "")
        f = files[i]
        siw = SentenceIteratorWrapper(f, 'wiki-sentences')
        it = siw.iterator(sentence_id)
        success, _ = bulk(es, it)
        sentence_id += siw.num_sentences_read
    es.indices.refresh(index="wiki-sentences")


if __name__ == "__main__":
    main(sys.argv[1:])
