import numpy
from keras.optimizers import Adam
from keras.models import Model
from refactored.self_attention import SelfAttention

## Copied from own keras fork.
from refactored.crf import ChainCRF

SEED = 0
# fix random seed for reproducibility
numpy.random.seed(SEED)

from keras.callbacks import *
from keras.layers import *

import numpy
from refactored import utils

def merged_input_layers(max_length, embedding_size, word_to_index, vocab_size, position_embedding_size, rellist):
    input_seq = Input(shape=(max_length,))
    prefix_seq = Input(shape=(max_length,))
    suffix_seq = Input(shape=(max_length,))
    position_seq = Input(shape=(max_length,))
    relation = Input(shape=(1,))

    embedding_file = "/mounts/Users/cisintern/beroth/workspace/relation_tagger/data/glove.6B.100d.txt"
    if embedding_file:
        emb_matrix = utils.init_embeddings(embedding_file, embedding_size, word_to_index)
        lookup = Embedding(input_dim=vocab_size, output_dim=embedding_size, weights=[emb_matrix])
    else:
        lookup = Embedding(input_dim=vocab_size, output_dim=embedding_size)

    word_embeddings = lookup(input_seq)
    prefix_embeddings = lookup(prefix_seq)
    suffix_embeddings = lookup(suffix_seq)

    position_embeddings = Embedding(input_dim=(2 * max_length), output_dim=position_embedding_size)(position_seq)

    relation_embedding_size = len(rellist)
    relation_embedding = Embedding(input_dim=len(rellist) + 1, output_dim=relation_embedding_size)(relation)
    relation_embedding = Flatten()(relation_embedding)
    relation_embedding_per_position = RepeatVector(max_length)(relation_embedding)

    merged_input_embeddings = merge([word_embeddings, prefix_embeddings, suffix_embeddings, position_embeddings,
                           relation_embedding_per_position], mode="concat")
    inputs = [input_seq, prefix_seq, suffix_seq, position_seq, relation]
    return inputs, merged_input_embeddings, relation_embedding_per_position

def encode_sequence_with_self_attention(max_length, word_to_index, vocab_size, rellist):
    embedding_size = 100
    num_attentions = 8
    hidden_size = 200
    position_embedding_size = 10
    hidden_layers = 3

    inputs, merged_inputs, _ = merged_input_layers(max_length, embedding_size, word_to_index, vocab_size,
                                                position_embedding_size, rellist)
    sequence = TimeDistributed(Dense(hidden_size, activation="relu"))(merged_inputs)
    #sequence =

    for i in range(hidden_layers):
        self_attended_seqs = []
        for j in range(num_attentions):
            self_attn = SelfAttention(hidden_size//num_attentions)(sequence)
            self_attended_seqs.append(self_attn)
            #self_attended_seqs.append(Dropout(0.2)(self_attn))
        self_attended_seq = Concatenate()(self_attended_seqs)
        #self_attended_seq = Add()([sequence, self_attended_seq])
        self_attended_seq = BatchNormalization()(self_attended_seq)
        #self_attended_seq = TimeDistributed(BatchNormalization())(self_attended_seq) # TODO: Layer Normalization
        feed_forward = TimeDistributed(Dense(2*hidden_size, activation='relu'))(self_attended_seq)
        feed_forward = TimeDistributed(Dense(hidden_size, activation='relu'))(feed_forward)
        # TODO: additional Dense layer
        #feed_forward = Add()([feed_forward, self_attended_seq])
        feed_forward = Add()([feed_forward, sequence])
        feed_forward = BatchNormalization()(feed_forward)
        sequence = feed_forward #TimeDistributed(BatchNormalization())(feed_forward) # TODO: Layer Normalization

    return inputs, sequence, hidden_size

def encode_sequence_with_rnn(max_length, word_to_index, vocab_size, rellist):
    embedding_size = 100
    hidden_size = 200
    position_embedding_size = 10
    hidden_layers = 2

    inputs, merged_inputs, _ = merged_input_layers(max_length, embedding_size, word_to_index, vocab_size,
                                                position_embedding_size, rellist)

    rnn_in_out = merged_inputs
    for i in range(hidden_layers):
        rnn_in_out = Bidirectional(GRU(hidden_size//2, return_sequences=True))(rnn_in_out)
    #context_vector = RepeatVector(self.max_length)(relation_embedding)
    #sequence = merge([rnn_in_out, context_vector], mode="concat")
    #for i in range(dense_layers):
    #    sequence = TimeDistributed(Dense(hidden_size, activation=final_hidden))(sequence)#W_regularizer=regularizers.l1_l2(l1_reg, l2_reg))(sequence))
    #    sequence = Dropout(hidden_dropout)(sequence)
    return inputs, rnn_in_out, hidden_size

def encode_sequence_with_rnn_ablation(max_length, word_to_index, vocab_size, rellist, merged_input_layers):
    embedding_size = 100
    hidden_size = 200
    position_embedding_size = 10
    hidden_layers = 2

    inputs, merged_inputs, _ = merged_input_layers(max_length, embedding_size, word_to_index, vocab_size,
                                                   position_embedding_size, rellist)

    rnn_in_out = merged_inputs
    for i in range(hidden_layers):
        rnn_in_out = Bidirectional(GRU(hidden_size//2, return_sequences=True))(rnn_in_out)
    #context_vector = RepeatVector(self.max_length)(relation_embedding)
    #sequence = merge([rnn_in_out, context_vector], mode="concat")
    #for i in range(dense_layers):
    #    sequence = TimeDistributed(Dense(hidden_size, activation=final_hidden))(sequence)#W_regularizer=regularizers.l1_l2(l1_reg, l2_reg))(sequence))
    #    sequence = Dropout(hidden_dropout)(sequence)
    return inputs, rnn_in_out, hidden_size


def encode_sequence_with_cnn(max_length, word_to_index, vocab_size, rellist):
    embedding_size = 100
    position_embedding_size = 10
    convolution_window_sizes = [3, 5, 7, 9]

    inputs, merged_embeddings, relation_embedding_per_position = \
        merged_input_layers(max_length, embedding_size, word_to_index, vocab_size, position_embedding_size, rellist)
    conv_dropout = 0.2

    cnn_layer1a = Convolution1D(nb_filter=32, filter_length=convolution_window_sizes[0], \
                                activation='relu', border_mode='same')(merged_embeddings)
    cnn_layer1b = Convolution1D(nb_filter=64, filter_length=convolution_window_sizes[0], \
                                activation='relu', border_mode='same')(Dropout(conv_dropout)(cnn_layer1a))
    cnn_layer1c = Convolution1D(nb_filter=128, filter_length=convolution_window_sizes[0], \
                                activation='relu', border_mode='same')(Dropout(conv_dropout)(cnn_layer1b))

    cnn_layer2a = Convolution1D(nb_filter=32, filter_length=convolution_window_sizes[1], \
                                activation='relu', border_mode='same')(merged_embeddings)
    cnn_layer2b = Convolution1D(nb_filter=64, filter_length=convolution_window_sizes[1], \
                                activation='relu', border_mode='same')(Dropout(conv_dropout)(cnn_layer2a))
    cnn_layer2c = Convolution1D(nb_filter=128, filter_length=convolution_window_sizes[1], \
                                activation='relu', border_mode='same')(Dropout(conv_dropout)(cnn_layer2b))

    cnn_layer3a = Convolution1D(nb_filter=32, filter_length=convolution_window_sizes[2], \
                                activation='relu', border_mode='same')(merged_embeddings)
    cnn_layer3b = Convolution1D(nb_filter=64, filter_length=convolution_window_sizes[2], \
                                activation='relu', border_mode='same')(Dropout(conv_dropout)(cnn_layer3a))
    cnn_layer3c = Convolution1D(nb_filter=128, filter_length=convolution_window_sizes[2], \
                                activation='relu', border_mode='same')(Dropout(conv_dropout)(cnn_layer3b))

    cnn_layer4a = Convolution1D(nb_filter=32, filter_length=convolution_window_sizes[3], \
                                activation='relu', border_mode='same')(merged_embeddings)
    cnn_layer4b = Convolution1D(nb_filter=64, filter_length=convolution_window_sizes[3], \
                                activation='relu', border_mode='same')(Dropout(conv_dropout)(cnn_layer4a))
    cnn_layer4c = Convolution1D(nb_filter=128, filter_length=convolution_window_sizes[3], \
                                activation='relu', border_mode='same')(Dropout(conv_dropout)(cnn_layer4b))

    sequence = merge([cnn_layer1c, cnn_layer2c, cnn_layer3c, cnn_layer4c, relation_embedding_per_position], mode="concat") # TODO: Test for table filling
    # sequence = merge([cnn_layer1c, cnn_layer2c, cnn_layer3c, cnn_layer4c], mode="concat") #

    sequence = Dropout(conv_dropout)(sequence)
    return inputs, sequence, 128 * 4 + len(rellist)

class PointerModel():
    def __init__(self,
                 trainpath,
                 devpath,
                 testpath,
                 vocabpath,
                 relfile="/mounts/Users/cisintern/beroth/data/tac/rellist2013"):
        self.epochs = 10
        self.batch_size = 512

        self.max_length = 40

        # maximum number of lines to be read from each set. If None, all lines are read
        self.maxlines = {"train": None, "dev": None, "test": None}

        self.datasets = {"train": trainpath, \
                         "dev": devpath, \
                         "test": testpath}

        self.labels = ("O", "I")  # target labels; I = inside the answer

        self.rellist = ["<UNK>"]
        for line in open(relfile):
            line = line.strip()
            if line:
                self.rellist.append(line)

        self.relation_to_index = {r:i for i,r in enumerate(self.rellist)}
        self.vocab_size = 100000
        self.num_relations = len(self.relation_to_index.keys())

        if trainpath and devpath:
            self.train_x, self.train_prefix, self.train_suffix, self.word_to_index, self.train_rel, \
            self.position_indices_train, self.train_label, self.train_start, self.train_end = \
                utils.load_word_seqs_prefix_suffix_position_file(trainpath, self.rellist, self.max_length,
                                                                 max_vocabsize=self.vocab_size, model_type=utils.POINTER_MODEL)

            self.valid_x, self.valid_prefix, self.valid_suffix, _, self.valid_rel, \
            self.position_indices_valid, self.valid_label, self.valid_start, self.valid_end = utils.load_word_seqs_prefix_suffix_position_file(
                devpath, self.rellist, self.max_length, self.word_to_index, max_vocabsize=self.vocab_size, model_type=utils.POINTER_MODEL)
            idx_to_vocab = {idx:token for token, idx in self.word_to_index.items()}
            whandle = open(vocabpath, "w")
            for i in range(len(idx_to_vocab)):
                whandle.write(idx_to_vocab[i] + "\n")
            whandle.close()
        else:
            self.word_to_index = dict()
            count = 0
            with open(vocabpath, "r") as handle:
                for line in handle:
                    self.word_to_index[line.rstrip('\n')] = count
                    count += 1


    def encode_sequence(self):
        raise NotImplementedError

    def model_entity_recognition(self, config):
        """
        Creates the model according to our parameters
        """
        final_hidden = 'relu'
        hidden_dropout = 0.2
        label_weight = 0.3
        position_weight = 1.0 - label_weight
        hidden_size = 200
        final_hidden = 'relu'
        dense_layers = 1
        optimizer = 'adam'
        learning_rate = 0.005
        decay = 1e-07

        inputs, hidden_seq, hidden_seq_dim = self.encode_sequence()

        def attention_weights(sequence, sequence_length, context_vectors=None, hidden_size=200, name=None):
            if context_vectors:
                if len(context_vectors) > 1:
                    context_vector = merge(context_vectors, mode="concat")
                else:
                    context_vector = context_vectors[0]
                context_vector = RepeatVector(sequence_length)(context_vector)
                sequence = merge([sequence, context_vector], mode="concat")
            for i in range(dense_layers):
                sequence = TimeDistributed(Dense(hidden_size, activation=final_hidden))(sequence)
                sequence = Dropout(hidden_dropout)(sequence)
            alpha = TimeDistributed(Dense(1))(sequence)
            alpha = Flatten()(alpha)
            return Activation(activation="softmax", name=name)(alpha)

        def weighted_average(sequence, weights, hidden_size):
            weights = RepeatVector(hidden_size)(weights)
            weights = Permute((2,1))(weights)
            weighted_sequence = merge([sequence, weights], mode='mul')# dot', dot_axes=1
            return GlobalAveragePooling1D()(weighted_sequence)

        use_attn_avg = False

        if use_attn_avg:
            label_attn_weights = attention_weights(hidden_seq, self.max_length, hidden_size=hidden_size)
            label_attn_vec = weighted_average(hidden_seq, label_attn_weights, hidden_seq_dim)

            start_attn_weights = attention_weights(hidden_seq, self.max_length, hidden_size=hidden_size, context_vectors=[label_attn_weights])
            start_attn_vec = weighted_average(hidden_seq, start_attn_weights, hidden_seq_dim)

            end_attn_weights = attention_weights(hidden_seq, self.max_length, hidden_size=hidden_size, context_vectors=[start_attn_vec])
            output_labels = Dense(1, activation='sigmoid', name="label")(label_attn_vec)
        else:
            pool = GlobalMaxPooling1D()(hidden_seq)
            intermediate_repr = pool
            #intermediate_repr = merge([intermediate_repr], mode='concat') # TODO: remove relation embedding as context?

            for i in range(dense_layers):
                intermediate_repr = Dense(hidden_size, activation=final_hidden)(intermediate_repr)
                intermediate_repr = Dropout(hidden_dropout)(intermediate_repr)
            #    output_labels = Dense(len(self.rellist)+1, activation='sigmoid', W_regularizer=regularizers.l1_l2(l1_reg, l2_reg), name="label")(penultimate_layer)
            output_labels = Dense(1, activation='sigmoid', name="label")(intermediate_repr)
            start_attn_weights = attention_weights(hidden_seq, self.max_length,
                                             context_vectors=[pool], hidden_size=hidden_size, name="start")
            end_attn_weights = attention_weights(hidden_seq, self.max_length,
                                           context_vectors=[start_attn_weights], hidden_size=hidden_size, name="len")

        output_start = start_attn_weights
        output_end = end_attn_weights

        if 'adam' == optimizer:
            optimizer_obj = Adam(lr=learning_rate, decay=decay)
        else:
            raise ValueError("Not a valid optimizer: " + optimizer)

        self.model = Model(input=inputs,
                           output=[output_labels, output_start, output_end])

        self.model.compile(loss=['binary_crossentropy','categorical_crossentropy','categorical_crossentropy'], optimizer=optimizer_obj, metrics=['accuracy'],
                           loss_weights=[label_weight, 0.5*position_weight, 0.5*position_weight])

    def predict_answer(self, sentence, relation):
        """
        Predict answer (as a string) given a sentence and a relation
        """
        sentence_tokens = sentence.strip().split()
        query_position = sentence_tokens.index(utils.QUERY)

        sentence_tokens = utils.fill_up_or_cut_right(sentence_tokens, utils.PAD_STRING, self.max_length)
        prefix_tokens = utils.prefixes(sentence_tokens)
        suffix_tokens = utils.suffixes(sentence_tokens)

        sentence_array = utils.tokens_to_indices_matrix(sentence_tokens, self.word_to_index)
        prefix_array = utils.tokens_to_indices_matrix(prefix_tokens, self.word_to_index)
        suffix_array = utils.tokens_to_indices_matrix(suffix_tokens, self.word_to_index)
        position_array = numpy.array([utils.position_indices(query_position, self.max_length)])
        relation_array = numpy.array([self.relation_to_index.get(relation, 0)])

        label_start_end = self.model.predict([sentence_array, prefix_array, suffix_array, position_array, relation_array])

        label_prob = label_start_end[0][0]
        label_pred = label_start_end[0][0] > 0.5

        start_prob = np.max(label_start_end[1][0])
        start_pred = np.argmax(label_start_end[1][0])

        end_prob = np.max(label_start_end[2][0])
        end_pred = np.argmax(label_start_end[2][0])

        score = (label_prob + start_prob + end_prob) / 3.0

        limit = min(len(sentence.strip().split()), self.max_length)

        if label_pred and start_pred < limit and (query_position < start_pred or end_pred < query_position):
            return " ".join(sentence_tokens[start_pred:end_pred+1]), score
        else:
            return "", 0.0

    def train_entity_recognition(self, modelpath):
        """
		Train the model on the training set, with early stopping on the dev set.
		Models are saved after every epoch.
		"""
        early_stopper = EarlyStopping(patience=1, monitor='val_start_acc', mode="max")
        checkpoint = ModelCheckpoint(modelpath, save_best_only=True, monitor='val_start_acc', mode="max")#
        self.model.fit([self.train_x, self.train_prefix, self.train_suffix, self.position_indices_train, self.train_rel],
                       [self.train_label, self.train_start, self.train_end], batch_size=self.batch_size,
                       callbacks=[early_stopper, checkpoint], nb_epoch=self.epochs,
                       validation_data=([self.valid_x, self.valid_prefix, self.valid_suffix,
                                         self.position_indices_valid, self.valid_rel],
                                        [self.valid_label, self.valid_start, self.valid_end]), \
                       verbose=1)
        self.model.load_weights(modelpath)

    def predict_dataset(self, dataset, filename=None):
        maxlines = self.maxlines[dataset]
        whandle = open(filename, "w")
        count = 0
        with open(self.datasets[dataset], "r") as handle:
            for line in handle:
                if (not maxlines is None) and count >= maxlines: break
                count += 1
                [sentence, query_or_ids, relation, answer] = line.strip().split("\t")
                predicted_answer, score = self.predict_answer(sentence, relation)
                predicted_answer = (predicted_answer + " </s>").strip()
                whandle.write("\t".join([sentence, query_or_ids, relation, predicted_answer]) + "\n")
        whandle.close()

    def predict_dataset_fast(self, dataset, filename_out=None, use_crf=False):
        """
        Predict answer (as a string) given a sentence and a relation
        """
        filename_in = self.datasets[dataset]
        test_x, test_prefix, test_suffix, _, test_rel, position_indices_test, test_label, test_start, test_end = \
            utils.load_word_seqs_prefix_suffix_position_file(filename_in, self.rellist, self.max_length,
                                                             self.word_to_index, max_vocabsize=self.vocab_size,
                                                             model_type=utils.POINTER_MODEL)
        data_input = [test_x, test_prefix, test_suffix, position_indices_test, test_rel]
        label_start_end = self.model.predict(data_input)
        whandle = open(filename_out, "w")
        instance_nr = 0
        with open(filename_in, "r") as handle:
            for line in handle:
                [sentence, query_or_ids, relation, answer] = line.strip().split("\t")
                sentence_tokens = sentence.strip().split()
                query_position = sentence_tokens.index(utils.QUERY)
                limit = min(len(sentence.strip().split()), self.max_length)
                label_prob = label_start_end[0][instance_nr]
                label_pred = label_start_end[0][instance_nr] > 0 # TODO: use 0.5?

                start_prob = np.max(label_start_end[1][instance_nr])
                start_pred = np.argmax(label_start_end[1][instance_nr])

                end_prob = np.max(label_start_end[2][instance_nr])
                end_pred = np.argmax(label_start_end[2][instance_nr])

                score = (label_prob + start_prob + end_prob) / 3.0

                if label_pred and start_pred < limit and (query_position < start_pred or end_pred < query_position):
                    predicted_answer = " ".join(sentence_tokens[start_pred:end_pred+1])
                else:
                    predicted_answer = ""
                    # score = 0.0
                instance_nr+=1
                whandle.write("\t".join([sentence, query_or_ids, relation, predicted_answer]) + "\n")
        whandle.close()

class TaggerModel():
    def __init__(self,
                 trainpath,
                 devpath,
                 testpath,
                 vocabpath,
                 relfile):
        print("Using crf with masking.")
        self.epochs = 10
        self.batch_size = 512

        self.max_length = 40

        # maximum number of lines to be read from each set. If None, all lines are read
        self.maxlines = {"train": None, "dev": None, "test": None}

        self.datasets = {"train": trainpath, \
                         "dev": devpath, \
                         "test": testpath}

        self.labels = ("O", "I")  # target labels; I = inside the answer

        # one hot vectors associated with the labels
        self.one_hot = {self.labels[idx]: [1 * idx == jdx for jdx in range(len(self.labels))] for idx in
                        range(len(self.labels))}

        self.rellist = ["<UNK>"]
        for line in open(relfile):
            line = line.strip()
            if line:
                self.rellist.append(line)

        self.relation_to_index = {r:i for i,r in enumerate(self.rellist)}
        self.vocab_size = 100000#len(self.word_to_index.keys())
        self.num_relations = len(self.relation_to_index.keys())

        if trainpath and devpath:
            self.train_x, self.train_prefix, self.train_suffix, self.word_to_index, self.train_rel, \
            self.position_indices_train, self.train_labels, self.train_sample_weights = \
                utils.load_word_seqs_prefix_suffix_position_file(trainpath, self.rellist, self.max_length,
                                                                 max_vocabsize=self.vocab_size, model_type=utils.TAGGING_MODEL)
            self.valid_x, self.valid_prefix, self.valid_suffix, _, self.valid_rel, \
            self.position_indices_valid, self.valid_labels, self.valid_sample_weights = \
                utils.load_word_seqs_prefix_suffix_position_file(devpath, self.rellist, self.max_length,
                                                                 self.word_to_index, max_vocabsize=self.vocab_size, model_type=utils.TAGGING_MODEL)
            idx_to_vocab = {idx:token for token, idx in self.word_to_index.items()}
            whandle = open(vocabpath, "w")
            for i in range(len(idx_to_vocab)):
                whandle.write(idx_to_vocab[i] + "\n")
            whandle.close()
        else:
            self.word_to_index = dict()
            count = 0
            with open(vocabpath, "r") as handle:
                for line in handle:
                    self.word_to_index[line.rstrip('\n')] = count
                    count += 1

    def encode_sequence(self):
        raise NotImplementedError

    def model_entity_recognition(self, config):
        """
		Creates the model according to our parameters
		"""
        #'batch_size': 512, , 'num_epochs': 1
        use_crf = config["crf"]
        optimizer = 'adam'
        learning_rate = 0.005
        decay = 1e-07

        inputs, hidden_seq, _ = self.encode_sequence()

        if use_crf:
            crf_scores_raw = TimeDistributed(Dense(2), name="crf_scores")(hidden_seq)
            crf = ChainCRF(name="crf")
            probabilities = crf(crf_scores_raw)
            loss = crf.loss
        else:
            probabilities = TimeDistributed(Dense(2, activation="softmax"))(hidden_seq)
            loss = "categorical_crossentropy"

        if 'adam' == optimizer:
            optimizer_obj = Adam(lr=learning_rate, decay=decay)
        else:
            raise ValueError("Not a valid optimizer: " + optimizer)

        self.model = Model(input=inputs, output=probabilities)
        self.model.compile(metrics=[self.p_i, self.r_i, self.f1_i], \
                           sample_weight_mode="temporal", \
                           loss=loss, optimizer=optimizer_obj)


    def pretty_print(self, token_ids, rel_ids, labels, sample_weights):
        idx_to_vocab = {idx:token for token, idx in self.word_to_index.items()}
        num_instances = len(token_ids)
        assert(num_instances == len(rel_ids))
        assert(num_instances == len(labels))
        assert(num_instances == len(sample_weights))
        for i in range(num_instances):
            tokens = [idx_to_vocab[tok_id] for tok_id in token_ids[i]]
            token_labels = [str(np.argmax(labelvec)) for labelvec in labels[i]]
            token_weights = [str(w) for w in sample_weights[i]]
            print(" ".join(["/".join([tokens[i], token_labels[i], token_weights[i]]) for i in range(len(tokens))]))

    # Metrics for evaluation
    # We defined our own precision, recall and F1 metrics in such a way that they only consider the 'I' label.
    # This is because the majority of correct labels are 'O', leading to high F1 scores even when the model
    # hardly predicts any 'I' labels
    def p_i(self, y_true, y_pred):
        """
		I-only precision
		"""
        true_positives_bi = K.sum(K.round(K.clip(K.variable(numpy.array(self.one_hot["I"])) * y_true * y_pred, 0, 1)))
        predicted_positives_bi = K.sum(K.round(K.clip(K.variable(numpy.array(self.one_hot["I"])) * y_pred, 0, 1)))
        return true_positives_bi / (predicted_positives_bi + K.epsilon())

    def r_i(self, y_true, y_pred):
        """
		I-only recall
		"""
        true_positives_bi = K.sum(K.round(K.clip(K.variable(numpy.array(self.one_hot["I"])) * y_true * y_pred, 0, 1)))
        possible_positives_bi = K.sum(K.round(K.clip(K.variable(numpy.array(self.one_hot["I"])) * y_true, 0, 1)))
        return true_positives_bi / (possible_positives_bi + K.epsilon())

    def f1_i(self, y_true, y_pred):
        """
		I-only F1
		"""
        recall = self.r_i(y_true, y_pred)
        precision = self.p_i(y_true, y_pred)
        return 2 * recall * precision / (recall + precision + K.epsilon())

    def predict_dataset_fast(self, dataset, filename_out=None, use_crf=False):
        """
		Predict answer (as a string) given a sentence and a relation
		"""
        filename_in = self.datasets[dataset]
        sentence_array, prefix_array, suffix_array, _, relation_array, \
        position_array, dummy_labels, dummy_sample_weights = \
            utils.load_word_seqs_prefix_suffix_position_file(filename_in, self.rellist, self.max_length,
                                                             self.word_to_index, max_vocabsize=self.vocab_size, model_type=utils.TAGGING_MODEL)

        data_input = [sentence_array, prefix_array, suffix_array, position_array, relation_array]
        predictions_ = self.model.predict(data_input)

        if use_crf:
            crf_unigram_model = Model(inputs=self.model.input,
                                      outputs=self.model.get_layer('crf_scores').output)
            unigram_predictions_ = crf_unigram_model.predict(data_input)

        def softmax(w):
            w = numpy.array(w)
            maxes = numpy.amax(w, axis=1, keepdims=True)
            e = numpy.exp(w - maxes)
            return e / numpy.sum(e, axis=1, keepdims=True)

        whandle = open(filename_out, "w")
        instance_nr = 0
        with open(filename_in, "r") as handle:
            for line in handle:
                predictions = predictions_[instance_nr]
                unigram_predictions = softmax(unigram_predictions_[instance_nr])
                [sentence, query_or_ids, relation, answer] = line.strip().split("\t")
                sentence_tokens = sentence.strip().split()
                limit = min(len(sentence.strip().split()), self.max_length)
                word_idx_list_of_lists = []
                scores = []
                #print()
                found_label = False
                for word_idx in range(limit):
                    predicted_label = predictions[word_idx].argmax()
                    #print(sentence_tokens[word_idx]+"/"+self.labels[predicted_label],end=" ")
                    if use_crf:
                        probability = unigram_predictions[word_idx,1]
                    else:
                        probability = predictions[word_idx].max()
                    if self.labels[predicted_label] == "I":
                        found_label = True
                        if len(word_idx_list_of_lists) == 0 or (not word_idx - 1 in word_idx_list_of_lists[-1]):
                            word_idx_list_of_lists.append([word_idx])
                            scores.append(probability)
                        else:
                            word_idx_list_of_lists[-1].append(word_idx)
                            scores[-1] += probability
                #if found_label:
                #    print(" ".join(sentence_tokens))
                #    print(" ".join([self.labels[predictions[word_idx].argmax()] for word_idx in range(limit)]))
                for list_nr in range(len(word_idx_list_of_lists)):
                    scores[list_nr] = scores[list_nr] / float(len(word_idx_list_of_lists[list_nr]))
                num_answers = len(word_idx_list_of_lists)
                if num_answers == 0:
                    predicted_answer = ""
                    score_of_predicted_answer = 0.0
                else:
                    list_indices = list(range(num_answers))
                    def score_function(x):
                        return scores[x]
                    sorted_results = sorted(list_indices, key=score_function, reverse=True)
                    def answer_string(result):
                        answer_indices = word_idx_list_of_lists[result]
                        return " ".join([sentence_tokens[i] for i in answer_indices])
                    predicted_answer = answer_string(sorted_results[0])
                    score_of_predicted_answer = scores[sorted_results[0]]
                predicted_answer = (predicted_answer + " </s>").strip()
                query = query_or_ids
                whandle.write("\t".join([sentence, query, relation, predicted_answer, str(score_of_predicted_answer)]) + "\n")
                instance_nr += 1
        whandle.close()


    def predict_dataset_fast_ablation(self, dataset, filename_out=None, use_crf=False, use_word=True, use_affix=True,
                                      use_position=True, use_relation=True):
        """
		Predict answer (as a string) given a sentence and a relation
		"""
        filename_in = self.datasets[dataset]
        sentence_array, prefix_array, suffix_array, _, relation_array, \
        position_array, dummy_labels, dummy_sample_weights = \
            utils.load_word_seqs_prefix_suffix_position_file(filename_in, self.rellist, self.max_length,
                                                             self.word_to_index, max_vocabsize=self.vocab_size, model_type=utils.TAGGING_MODEL)

        data_input = [] #[sentence_array, prefix_array, suffix_array, position_array, relation_array]
        if use_word:
            data_input.append(sentence_array)
        if use_affix:
            data_input.extend([prefix_array, suffix_array])
        if use_position:
            data_input.extend(position_array)
        if use_relation:
            data_input.extend(relation_array)

        predictions_ = self.model.predict(data_input)

        if use_crf:
            crf_unigram_model = Model(inputs=self.model.input,
                                      outputs=self.model.get_layer('crf_scores').output)
            unigram_predictions_ = crf_unigram_model.predict(data_input)

        def softmax(w):
            w = numpy.array(w)
            maxes = numpy.amax(w, axis=1, keepdims=True)
            e = numpy.exp(w - maxes)
            return e / numpy.sum(e, axis=1, keepdims=True)

        whandle = open(filename_out, "w")
        instance_nr = 0
        with open(filename_in, "r") as handle:
            for line in handle:
                predictions = predictions_[instance_nr]
                unigram_predictions = softmax(unigram_predictions_[instance_nr])
                [sentence, query_or_ids, relation, answer] = line.strip().split("\t")
                sentence_tokens = sentence.strip().split()
                limit = min(len(sentence.strip().split()), self.max_length)
                word_idx_list_of_lists = []
                scores = []
                #print()
                found_label = False
                for word_idx in range(limit):
                    predicted_label = predictions[word_idx].argmax()
                    #print(sentence_tokens[word_idx]+"/"+self.labels[predicted_label],end=" ")
                    if use_crf:
                        probability = unigram_predictions[word_idx,1]
                    else:
                        probability = predictions[word_idx].max()
                    if self.labels[predicted_label] == "I":
                        found_label = True
                        if len(word_idx_list_of_lists) == 0 or (not word_idx - 1 in word_idx_list_of_lists[-1]):
                            word_idx_list_of_lists.append([word_idx])
                            scores.append(probability)
                        else:
                            word_idx_list_of_lists[-1].append(word_idx)
                            scores[-1] += probability
                #if found_label:
                #    print(" ".join(sentence_tokens))
                #    print(" ".join([self.labels[predictions[word_idx].argmax()] for word_idx in range(limit)]))
                for list_nr in range(len(word_idx_list_of_lists)):
                    scores[list_nr] = scores[list_nr] / float(len(word_idx_list_of_lists[list_nr]))
                num_answers = len(word_idx_list_of_lists)
                if num_answers == 0:
                    predicted_answer = ""
                    score_of_predicted_answer = 0.0
                else:
                    list_indices = list(range(num_answers))
                    def score_function(x):
                        return scores[x]
                    sorted_results = sorted(list_indices, key=score_function, reverse=True)
                    def answer_string(result):
                        answer_indices = word_idx_list_of_lists[result]
                        return " ".join([sentence_tokens[i] for i in answer_indices])
                    predicted_answer = answer_string(sorted_results[0])
                    score_of_predicted_answer = scores[sorted_results[0]]
                predicted_answer = (predicted_answer + " </s>").strip()
                query = query_or_ids
                whandle.write("\t".join([sentence, query, relation, predicted_answer, str(score_of_predicted_answer)]) + "\n")
                instance_nr += 1
        whandle.close()

    def train_entity_recognition(self, modelpath):
        """
		Train the model on the training set, with early stopping on the dev set.
		Models are saved after every epoch.
		"""
        early_stopper = EarlyStopping(monitor='val_f1_i', patience=1, mode="max")
        checkpoint = ModelCheckpoint(modelpath,save_best_only=True,monitor='val_f1_i', mode="max")
        self.model.fit([self.train_x, self.train_prefix, self.train_suffix, self.position_indices_train, self.train_rel], [self.train_labels], batch_size=self.batch_size,
                       sample_weight=[self.train_sample_weights], \
                       callbacks=[early_stopper, checkpoint], nb_epoch=self.epochs,
                       validation_data=(
                           [self.valid_x, self.valid_prefix, self.valid_suffix, self.position_indices_valid, self.valid_rel],
                           [self.valid_labels], [self.valid_sample_weights]), \
                       verbose=1)
        self.model.load_weights(modelpath)

    def train_entity_recognition_ablation(self, modelpath, use_word=True, use_affix=True, use_position=True, use_relation=True):
        """
        Train the model on the training set, with early stopping on the dev set.
        Models are saved after every epoch.
        """
        early_stopper = EarlyStopping(monitor='val_f1_i', patience=1, mode="max")
        checkpoint = ModelCheckpoint(modelpath,save_best_only=True,monitor='val_f1_i', mode="max")
        train_input_list = [] # [self.train_x, self.train_prefix, self.train_suffix, self.position_indices_train, self.train_rel]
        valid_input_list = [] # [self.valid_x, self.valid_prefix, self.valid_suffix, self.position_indices_valid, self.valid_rel]
        if use_word:
            train_input_list.append(self.train_x)
            valid_input_list.append(self.valid_x)
        if use_affix:
            train_input_list.extend([self.train_prefix, self.train_suffix])
            valid_input_list.extend([self.valid_prefix, self.valid_suffix])
        if use_position:
            train_input_list.append(self.position_indices_train)
            valid_input_list.append(self.position_indices_valid)
        if use_relation:
            train_input_list.append(self.train_rel)
            valid_input_list.append(self.valid_rel)

        self.model.fit(train_input_list, [self.train_labels], batch_size=self.batch_size,
                       sample_weight=[self.train_sample_weights], \
                       callbacks=[early_stopper, checkpoint], nb_epoch=self.epochs,
                       validation_data=(
                           valid_input_list,
                           [self.valid_labels], [self.valid_sample_weights]), \
                       verbose=1)
        self.model.load_weights(modelpath)

class TableFillingModel():
    def __init__(self,
                 trainpath,
                 devpath,
                 testpath,
                 vocabpath,
                 relfile):

        self.rellist = ["<UNK>"]
        for line in open(relfile):
            line = line.strip()
            if line:
                self.rellist.append(line)
        self.epochs = 20
        self.batch_size = 512

        self.max_length = 40

        self.datasets = {"train": trainpath, \
                         "dev": devpath, \
                         "test": testpath}

        self.vocab_size = 100000 #100000#len(self.word_to_index.keys())

        self.labels = ("O", "I")  # TODO: replace with relation names
        # TODO: when learning:
        #class_weight = {0 : 1.,
        #                1: 50.,
        #                2: 2.}
        # TODO: or
        #sample_weight=[self.train_sample_weights], \

        # one hot vectors associated with the labels
        self.one_hot = {self.labels[idx]: [1 * idx == jdx for jdx in range(len(self.labels))] for idx in
                        range(len(self.labels))}
        self.max_answer_length = 5
        self.positive_class_weight = 2
        if trainpath and devpath:
            print("Reading training data ...")
            self.train_x, self.train_prefix, self.train_suffix, self.word_to_index, self.train_rel, \
            self.position_indices_train, self.train_labels, self.train_sample_weights = \
                utils.load_word_seqs_prefix_suffix_position_file(trainpath, self.rellist, self.max_length,
                                                                 max_vocabsize=self.vocab_size,
                                                                 model_type=utils.TABLE_MODEL,
                                                                 max_answer_length=self.max_answer_length,
                                                                 positive_class_weight=self.positive_class_weight)
            self.valid_x, self.valid_prefix, self.valid_suffix, _, self.valid_rel, \
            self.position_indices_valid, self.valid_labels, self.valid_sample_weights = \
                utils.load_word_seqs_prefix_suffix_position_file(devpath, self.rellist, self.max_length,
                                                                 self.word_to_index, max_vocabsize=self.vocab_size,
                                                                 model_type=utils.TABLE_MODEL,
                                                                 max_answer_length=self.max_answer_length,
                                                                 positive_class_weight=self.positive_class_weight)
            idx_to_vocab = {idx:token for token, idx in self.word_to_index.items()}
            whandle = open(vocabpath, "w")
            for i in range(len(idx_to_vocab)):
                whandle.write(idx_to_vocab[i] + "\n")
            whandle.close()
            print("... reading training data DONE!")
        else:
            self.word_to_index = dict()
            count = 0
            with open(vocabpath, "r") as handle:
                for line in handle:
                    self.word_to_index[line.rstrip('\n')] = count
                    count += 1
        self.vocab_size = len(self.word_to_index)

    def encode_sequence(self):
        raise NotImplementedError

    def model_entity_recognition(self, config):
        """
        Creates the model according to our parameters
        """
        #'batch_size': 512, , 'num_epochs': 1
        hidden_size = 200 #100 #200
        dense_layers = 0
        dense_activation = 'relu'

        optimizer = 'adam'
        learning_rate = 0.005
        decay = 1e-07

        print("Building model ...")
        inputs, hidden_seq, hidden_seq_dim = self.encode_sequence()

        sequence_of_starts = hidden_seq
        sequence_of_ends = None

        shift_left_kernel = np.asarray([np.eye(hidden_seq_dim),
                                        np.zeros((hidden_seq_dim,hidden_seq_dim)),
                                        np.zeros((hidden_seq_dim,hidden_seq_dim))])
        shift_left_conv = Convolution1D(hidden_seq_dim, 3, kernel_initializer="zeros", trainable=False, padding="same", weights=[shift_left_kernel], use_bias=False)

        probabilities = []
        for answer_length in range(self.max_answer_length):
            if sequence_of_ends is not None:
                sequence = merge([sequence_of_starts, sequence_of_ends], mode="concat")
            else:
                sequence = sequence_of_starts
                sequence_of_ends = sequence_of_starts
            for i in range(dense_layers):
                #sequence = Dropout(dense_dropout)(sequence)
                sequence = TimeDistributed(Dense(hidden_size, activation=dense_activation))(sequence)#W_regularizer=regularizers.l1_l2(l1_reg, l2_reg))(sequence))
            predicted = TimeDistributed(Dense(2, activation="softmax"), name="length_"+str(answer_length+1))(sequence)
            probabilities.append(predicted)
            sequence_of_ends = shift_left_conv(sequence_of_ends)
        loss = "categorical_crossentropy"
        if 'adam' == optimizer:
            optimizer_obj = Adam(lr=learning_rate, decay=decay)
        else:
            raise ValueError("Not a valid optimizer: " + optimizer)
        self.model = Model(input=inputs, output=probabilities)
        self.model.compile(metrics=[self.p_i, self.r_i, self.f1_i], loss=loss, optimizer=optimizer_obj, sample_weight_mode="temporal")
        print("... building model DONE!")

    def p_i(self, y_true, y_pred):
        """
        I-only precision
        """
        true_positives_bi = K.sum(K.round(K.clip(K.variable(numpy.array(self.one_hot["I"])) * y_true * y_pred, 0, 1)))
        predicted_positives_bi = K.sum(K.round(K.clip(K.variable(numpy.array(self.one_hot["I"])) * y_pred, 0, 1)))
        return (true_positives_bi + K.epsilon()) / (predicted_positives_bi + K.epsilon())

    def r_i(self, y_true, y_pred):
        """
		I-only recall
		"""
        true_positives_bi = K.sum(K.round(K.clip(K.variable(numpy.array(self.one_hot["I"])) * y_true * y_pred, 0, 1)))
        possible_positives_bi = K.sum(K.round(K.clip(K.variable(numpy.array(self.one_hot["I"])) * y_true, 0, 1)))
        return (true_positives_bi + K.epsilon()) / (possible_positives_bi + K.epsilon())

    def f1_i(self, y_true, y_pred):
        """
		I-only F1
		"""
        recall = self.r_i(y_true, y_pred)
        precision = self.p_i(y_true, y_pred)
        return 2 * recall * precision / (recall + precision + K.epsilon())

    def train_entity_recognition(self, modelpath):
        """
		Train the model on the training set, with early stopping on the dev set.
		Models are saved after every epoch.
		"""
        print("Training model ...")
        early_stopper = EarlyStopping(monitor='val_length_2_f1_i',patience=1, mode="max")
        checkpoint = ModelCheckpoint(modelpath,save_best_only=True,monitor='val_length_2_f1_i', mode="max")
        self.model.fit([self.train_x, self.train_prefix, self.train_suffix, self.position_indices_train, self.train_rel],
                       self.train_labels, batch_size=self.batch_size,
                       sample_weight=self.train_sample_weights, \
                       callbacks=[early_stopper, checkpoint], nb_epoch=self.epochs,
                       validation_data=(
                           [self.valid_x, self.valid_prefix, self.valid_suffix, self.position_indices_valid, self.valid_rel],
                           self.valid_labels, self.valid_sample_weights), \
                       verbose=1)
        print("... training model DONE!")
        self.model.load_weights(modelpath)

    def predict_dataset_fast(self, dataset, filename_out=None, tac_format=False):
        """
        Predict answer (as a string) given a sentence and a relation
        """
        filename_in = self.datasets[dataset]
        sentence_array, prefix_array, suffix_array, _, relation_array, \
        position_array, dummy_labels, dummy_sample_weights = \
            utils.load_word_seqs_prefix_suffix_position_file(filename_in, self.rellist, self.max_length,
                                                             self.word_to_index, max_vocabsize=self.vocab_size,
                                                             model_type=utils.TABLE_MODEL,
                                                             max_answer_length=self.max_answer_length,
                                                             positive_class_weight=self.positive_class_weight)
        data_input = [sentence_array, prefix_array, suffix_array, position_array, relation_array]
        predictions = self.model.predict(data_input)

        whandle = open(filename_out, "w")
        instance_nr = 0
        with open(filename_in, "r") as handle:
            for line in handle:
                [sentence, query_or_ids, relation, answer] = line.strip().split("\t")
                sentence_tokens = sentence.strip().split()
                limit = min(len(sentence.strip().split()), self.max_length)

                predicted_answers = []
                scores = []
                starts = set()
                ends = set()

                for answer_length in reversed(range(self.max_answer_length)):
                    #predicted_score_start_end = []
                    for start in range(limit):
                        score = predictions[answer_length][instance_nr][start,1]
                        end = start+answer_length+1
                        if score > 0.5 and start not in starts and end not in ends:
                            answer = " ".join(sentence_tokens[start:end])
                            predicted_answers.append(answer)
                            scores.append(score)
                            starts.add(start)
                            ends.add(end)
                if len(predicted_answers) == 0:
                    predicted_answer = ""
                else:
                    best_answer, best_score = "", 0
                    for predicted_answer, score in zip(predicted_answers, scores):
                        if score > best_score:
                            best_answer = predicted_answer
                            best_score = score
                    predicted_answer = best_answer
                predicted_answer = (predicted_answer + " </s>").strip()
                query = query_or_ids
                whandle.write("\t".join([sentence, query, relation, predicted_answer]) + "\n")
                instance_nr += 1
        whandle.close()