import gzip
from collections import Counter
import numpy as np
from keras.utils.np_utils import to_categorical
from keras.utils import np_utils

UNKNOWN_STRING = "<UNK>"
PAD_STRING = "_"
QUERY = "<QUERY>"



def configure_tensorflow():
    import keras.backend as K
    import tensorflow as tf
    import os
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"   # see issue #152
    #os.environ["CUDA_VISIBLE_DEVICES"] = "7"

    gpu_options = tf.GPUOptions(per_process_gpu_memory_fraction=0.9)
    gpu_config = tf.ConfigProto(gpu_options=gpu_options)
    session = tf.Session(config=gpu_config)
    K.set_session(session)


def init_embeddings(embedding_file, embedding_size, word_to_index):
    print("Initializing with pre-trained embeddings: " + embedding_file)
    emb_matrix = np.random.uniform(-0.05,0.05,(len(word_to_index), embedding_size))
    num_updated = 0
    for line in open(embedding_file):
        parts = line.split(" ")
        word = parts[0]
        if word in word_to_index:
            idx = word_to_index[word]
            vec = np.asarray([float(s) for s in parts[1:]])
            emb_matrix[idx] = vec
            num_updated+=1
    print("Updated %d out of %d entries" % (num_updated, len(word_to_index)))
    return emb_matrix

def to_matrix(list_of_lists, max_length, new_label=0):
    """This return a matrix of size len(list_of_lists) x max_length. lists_of_lists typically contains integer indices.
    Lists longer than max_length are cut off, for shorter lists the matrix is padded with 0.
    A new label can be passed - then all cells that were already assigned 0 in the original lists are replaced with that
    new label.
    """
    m = new_label * np.ones((len(list_of_lists), max_length), dtype=np.int)
    for i,l in enumerate(list_of_lists):
        l = l[:max_length]
        m[i,:len(l)]=l
    if new_label != 0:
        # Flip values so that padded cells become zero.
        is_zero = m==0
        padded = m==new_label
        m[is_zero] = new_label
        m[padded] = 0
    return m

def dict_for_topn(list, max_vocabsize, include_additionally=[UNKNOWN_STRING]):
    d = {s:i for i,s in enumerate([s for s,_ in Counter(list).most_common(max_vocabsize - len(include_additionally))])}
    for s in include_additionally:
        additional_id = d.get(s, len(d))
        d[s] = additional_id
    return d

def flatten(nested):
    return [item for sublist in nested for item in sublist]

def map_nested(nested, d, default_index):
    return [[d.get(item, default_index) for item in sublist] for sublist in nested]

def string_to_ids(list_of_lists, max_vocabsize, str_to_id=None):
    if not str_to_id:
        str_to_id = dict_for_topn(flatten(list_of_lists), max_vocabsize)
    id_lists = map_nested(list_of_lists, str_to_id, str_to_id.get(UNKNOWN_STRING))
    return id_lists, str_to_id

def fill_up_or_cut_right(string_list, filler_str, desired_size):
    res_list = string_list[:desired_size]
    res_list = [res_list[i] if i < len(string_list) else filler_str for i in range(desired_size)]
    return res_list

def sublist_index(lst, sublst, close_to_position=0):
    if len(sublst) == 0:
        return -1
    positions = range(len(lst) - len(sublst) + 1)
    match_position = -1
    closest_match_distance = len(lst) + 1
    for i in positions:
        if lst[i] == sublst[0] and lst[i:i+len(sublst)] == sublst:
            curr_distance = min(abs(i-close_to_position), abs(i+len(sublst) - close_to_position))
            if curr_distance < closest_match_distance:
                closest_match_distance = curr_distance
                match_position = i
    return match_position

def prefixes(tokens, affix_length=2):
    return ["_" + w[:affix_length] for w in tokens]

def suffixes(tokens, affix_length=2):
    return [w[-affix_length:] + "_" for w in tokens]

def position_indices(query_position, max_length):
    query_position = min(max_length - 1, query_position)
    start_pos_index = max_length - query_position - 1
    end_pos_index = start_pos_index + max_length
    return list(range(start_pos_index, end_pos_index))

def to_categorical_tensor_1label(idx_matrix,num_classes):
    shape0, shape1 = idx_matrix.shape
    assert(shape1==1)
    categorical = to_categorical(idx_matrix.flatten(), num_classes)
    return categorical.reshape(shape0, num_classes)

def encode_single_sentence(sentence, max_length, tokens_to_indices_matrix, relation_id):
    sentence_tokens = sentence.strip().split()
    query_position = sentence_tokens.index(QUERY)

    sentence_tokens = fill_up_or_cut_right(sentence_tokens, PAD_STRING, max_length)
    prefix_tokens = prefixes(sentence_tokens)
    suffix_tokens = suffixes(sentence_tokens)

    sentence_array = tokens_to_indices_matrix(sentence_tokens)
    prefix_array = tokens_to_indices_matrix(prefix_tokens)
    suffix_array = tokens_to_indices_matrix(suffix_tokens)
    position_array = np.array([position_indices(query_position, max_length)])
    relation_array = np.array([relation_id]) #relation_to_index.get(relation, 0)

    return sentence_array, prefix_array, suffix_array, position_array, relation_array

POINTER_MODEL = "POINTER_MODEL"
TAGGING_MODEL = "TAGGING_MODEL"
TABLE_MODEL = "TABLE_MODEL"

#TODO
label_subsequences=False

def encode_word_prefix_suffix(lines, rellist, max_length, word_dict=None, max_vocabsize = 100000, model_type=None,
                              max_answer_length=5, positive_class_weight=1):
    rel_to_id = {r:i for i,r in enumerate(rellist)}
    word_lists = []
    prefix_lists = []
    suffix_lists = []
    relation_list = []
    position_indices_list = []
    if model_type == POINTER_MODEL:
        zero_one_label_list = []
        answer_start_list = []
        answer_end_list = []
    elif model_type == TAGGING_MODEL:
        sample_weights_list = []
        labels_list = []
    elif model_type == TABLE_MODEL:
        sample_weights_lists = [[] for _ in range(max_answer_length)]
        label_lists_per_answer_length = [[] for _ in range(max_answer_length)]
    else:
        raise NotImplementedError("Unknown model type: " + model_type)

    for line in lines:
        parts = line.strip().split("\t")
        assert(len(parts) == 4)
        sentence_tokens = parts[0].split(" ")
        rel = parts[2]
        if rel not in rel_to_id.keys():
            raise ValueError("Unknown relation: " + rel)
        answer_tokens = parts[3].split(" ")[:-1] # remove </s>
        answer_length = len(answer_tokens)
        query_position = sentence_tokens.index(QUERY)
        answer_position = sublist_index(sentence_tokens, answer_tokens, close_to_position=query_position)
        sentence_tokens = fill_up_or_cut_right(sentence_tokens, PAD_STRING, max_length)
        position_indices_list.append(position_indices(query_position, max_length))
        word_lists.append(sentence_tokens)
        prefix_lists.append(prefixes(sentence_tokens))
        suffix_lists.append(suffixes(sentence_tokens))
        relation_list.append(rel_to_id[rel])
        if model_type == POINTER_MODEL:
            query_position = min(max_length - 1, query_position)
            # No answer in relevant window
            if answer_length == 0 or answer_position >= max_length or answer_position == -1:
                answer_start = query_position
                answer_end = answer_start
                label = 0
            else:
                answer_start = answer_position
                answer_end = min(max_length-1, answer_start + answer_length - 1)
                label = 1
            zero_one_label_list.append(label)
            answer_start_list.append([answer_start])
            answer_end_list.append([answer_end])
        elif model_type == TAGGING_MODEL:
            labels = [0] * max_length
            if answer_length > 0:
                for i in range(answer_position, min(answer_position + answer_length, max_length)):
                    labels[i] = 1
            labels = np_utils.to_categorical(labels,2)
            sample_weights = [1 if index < len(sentence_tokens) else 0 for index in range(max_length)]

            if (labels.shape != (max_length,2)):
                raise ValueError("Unexpected label shape" + str(labels.shape))
            labels_list.append(labels)
            sample_weights_list.append(sample_weights)
        elif model_type == TABLE_MODEL:
            binary_table = False
            if binary_table:
                curr_label_lists = [[[0]] * max_length for _ in range(max_answer_length)]
            else:
                curr_label_lists = [[0] * max_length for _ in range(max_answer_length)]
            if query_position < max_length and \
                            answer_length != 0 and \
                                    answer_position + answer_length <= max_length and \
                            answer_position != -1:
                answer_length = min(answer_length, max_answer_length)
                if label_subsequences:
                    for sublength in range(answer_length):
                        start_positions_for_length = curr_label_lists[sublength]
                        for answer_shift in range(answer_length - sublength):
                            start_positions_for_length[answer_position + answer_shift] = [1]
                else:
                    curr_label_lists[answer_length - 1][answer_position] = [1] if binary_table else 1
            for i in range(max_answer_length):
                if binary_table:
                    label_lists_per_answer_length[i].append(curr_label_lists[i]) # TODO: dont use one-hot encoding
                else:
                    label_lists_per_answer_length[i].append(np_utils.to_categorical(curr_label_lists[i],2))
                #curr_sample_weights = [1 if index < effective_length else 0 for index in range(max_length)]
                curr_sample_weights = [0]*max_length
                for j in range(min(len(sentence_tokens), max_length)-i):
                    curr_sample_weights[j] = 1 + (positive_class_weight - 1) * curr_label_lists[i][j] # TODO: 101 x more weight if positive
                sample_weights_lists[i].append(curr_sample_weights)
        else:
            raise NotImplementedError("Unknown model type: " + model_type)

    word_prefix_suffix_lists = word_lists + prefix_lists + suffix_lists
    char_mapped, word_dict = string_to_ids(word_prefix_suffix_lists, max_vocabsize, word_dict)
    m = to_matrix(char_mapped, max_length)
    num_inst = len(word_lists)
    m_word = m[:num_inst]
    m_prefix = m[num_inst:2*num_inst]
    m_suffix = m[2*num_inst:3*num_inst]
    m_position_indices = np.array(position_indices_list)
    if model_type == POINTER_MODEL:
        m_start = to_categorical_tensor_1label(to_matrix(answer_start_list, 1), max_length)
        m_end = to_categorical_tensor_1label(to_matrix(answer_end_list, 1), max_length)
        m_label =  np.array(zero_one_label_list)
        return (m_word, m_prefix, m_suffix, word_dict, np.array(relation_list), m_position_indices, m_label, m_start, m_end)
    elif model_type == TAGGING_MODEL:
        m_label = np.array(labels_list)
        m_sample_weights = np.array(sample_weights_list)
        return (m_word, m_prefix, m_suffix, word_dict, np.array(relation_list), m_position_indices, m_label, m_sample_weights)
    elif model_type == TABLE_MODEL:
        label_matrices =  [np.array(label_list) for label_list in label_lists_per_answer_length]
        sample_weights = [np.array(sw) for sw in sample_weights_lists]
        print(len(label_matrices))
        print(len(sample_weights))
        print("######")
        return (m_word, m_prefix, m_suffix, word_dict, np.array(relation_list), m_position_indices, label_matrices, sample_weights)
    else:
        raise NotImplementedError("Unknown model type: " + model_type)

def load_word_seqs_prefix_suffix_position_file(filename, rellist, max_length = 40, word_dict=None, max_vocabsize = 100000, model_type=None, max_answer_length=5, positive_class_weight=1):
    if filename.endswith(".gz"):
        f = gzip.open(filename, "rt")
    else:
        f = open(filename,"r")
    results = encode_word_prefix_suffix(f, rellist, max_length, word_dict, max_vocabsize=max_vocabsize,
                                        model_type=model_type, max_answer_length=max_answer_length,
                                        positive_class_weight=positive_class_weight)
    f.close()
    return results

def tokens_to_indices_matrix(tokens, word_to_index):
    """
    Transform a sentence into a list of indices
    """
    unk_idx = word_to_index[UNKNOWN_STRING]
    return np.array([[word_to_index.get(word, unk_idx) for word in tokens]])
